static const int EOL = -1;

#include <cctype>

#include "encoders.hh"
#include "csetdata.c"

class UTF8: public Encoding
{
public:
    virtual const Fondness GetFondness() const
    {
        return VeryReliable;
    }

    virtual void SeqValue(std::string& result, wchar_t ch, int&) const
    {
        const unsigned n = ch;
        if(n < 0x80)                 // <=7 bits
            result += (char)n;
        else
        {
            if(n < 0x800)            // <=11 bits
                result += (char)(0xC0 + (n>>6));
            else
            {
                if(n < 0x10000)      // <=16 bits
                    result += (char)(0xE0 + (n>>12));
                else                 // <=21 bits
                {
                    result += (char)(0xF0 + (n>>18));
                    result += (char)(0x80 + ((n>>12)&63));
                }
                result += (char)(0x80 + ((n>>6)&63));
            }
            result += (char)(0x80 + (n&63));
        }
    }
    virtual void SeqEnd(std::string&) const
    {
        // nothing to do
    }
    virtual bool Encodable(wchar_t ch, int shift) const
    {
        /* UTF-8 can encode practically any Unicode value. */
        return ch < 0x200000;
    }
    
    /* Reading */
    virtual const SeqResult IsShift(const std::string&, unsigned) const
    {
        return false;
    }
    virtual const SeqResult IsData(const std::string& input, unsigned pos, int) const
    {
        /* This must be the most obfuscated UTF-8 decoder I have seen.
         * I tried to avoid having many hardcoded magical numbers. -Bisqwit
         */
        unsigned char headbyte = input[pos];
        
        /* How many bytes does each sequence take? */
        static const char sizes[16] =
        { 1,1,1,1,1,1,1,1,
          0,0,0,0,2,2,3,4 };
        
        /* What is the minimum value of the given sequence? */
        static const unsigned minimums[4] = { 0, 0x80, 0x800, 0x10000 };
        /* How to mask bits from the first byte */
        static const char     masks[4]    = { 0x7F, 0x1F, 0x0F, 0x07 };
        
        unsigned len = sizes[headbyte >> 4];
        
        //fprintf(stderr, "UTF-8 in: %c -> len=%u\n", (char)headbyte, len);
        
        /* Ensure it wasn't an invalid sequence */
        if(len == 0) return false;
        /* Ensure we have enough bytes */
        if(pos+len > input.size()) return false;
        
        unsigned result=0, shl=0;
        
        /* Process the tail bytes - last to first. */
        for(unsigned n = len; --n > 0; )
        {
            unsigned char byte = input[pos+n];
            
            //fprintf(stderr, "byte %u: %02X\n", n, byte);
            
            // The tail bytes must be 10xxxxxx
            if((byte & 0xC0) != 0x80) return false;
            
            unsigned bits = byte & 0x3F;
            result |= bits << shl;
            shl += 6;
        }
        
        /* Process the head byte. */
        // The top bits have already been verified in sizes[].
        // Take the low bits.
        unsigned bits = headbyte & masks[len-1];
        result |= bits << shl;
        
        if(result < minimums[len-1])
        {
            /* Non-optimal coding - not likely valid. */
            return false;
        }

        //fprintf(stderr, "UTF-8 in: %.*s -> %X\n", len, input.c_str()+pos, result);
        
        return SeqResult(true, len, result);
    }
};

class LatinEncoding: public Encoding
{  
    unsigned short min_uni;
    unsigned short max_uni;
    const unsigned char* uni_iso;
    const unsigned short* iso_uni;

public:
    LatinEncoding(const unsigned char* ui,
                  const unsigned short* iu,
                  int min,
                  int max)
      : min_uni(min),
        max_uni(max),
        uni_iso(ui),
        iso_uni(iu)
    {
    }
    virtual ~LatinEncoding() {}

    virtual const Fondness GetFondness() const
    {
        return OftenGood;
    }

    virtual void SeqValue(std::string& result, wchar_t ch, int&) const
    {
        unsigned isovalue = ch;
        if(ch >= 0xA1)
        {
            unsigned unicode = ch;
            if(unicode >= min_uni && unicode <= max_uni)
                isovalue = uni_iso[unicode-min_uni];
        }
        result += (char)isovalue;
    }
    virtual void SeqEnd(std::string&) const
    {
        // nothing to do
    }
    virtual bool Encodable(wchar_t ch, int shift) const
    {
        if(ch < 0xA1) return true;
        unsigned unicode = ch;
        if(unicode >= min_uni && unicode <= max_uni)
            return uni_iso[unicode-min_uni] != 0;
        return false;
    }
    
    /* Reading */
    virtual const SeqResult IsShift(const std::string&, unsigned) const
    {
        return false;
    }
    virtual const SeqResult IsData(const std::string& input, unsigned pos, int) const
    {
        unsigned char byte = input[pos];
        
        if(byte < 0x7F)
        {
            return SeqResult(true, 1, byte);
        }
        if(byte < 0xA0)
        {
            /* For control characters, decoding is not really preferred,
             * but it's a valid last option, so we return this instead
             * of just "false".
             */
            return SeqResult(false, 1, byte);
        }
        if(byte < 0xA1)
        {
            return SeqResult(true, 1, byte);
        }
        return SeqResult(true, 1, iso_uni[byte - 0xA1]);
    }
};

class Iso2022JP: public Encoding
{
public:
    virtual ~Iso2022JP() {}
    virtual const Fondness GetFondness() const
    {
        return VeryReliable;
    }

    virtual void SeqValue(std::string& result, wchar_t ch, int& shift) const
    {
        int wanted_shift = 0;
        
        unsigned jisvalue=0, bytes=0;
        unsigned unicode = ch;
        
        if(unicode >= min_uni_euc && unicode <= max_uni_euc)
        {
            jisvalue = uni_euc[unicode-min_uni_euc];
            if(jisvalue != 0)
            {
                wanted_shift = 1;
                bytes = 2;
            }
        }
        if(jisvalue == 0
        && unicode >= 0xFF61 && unicode <= 0xFF9F) // thin kana
        {
            jisvalue = unicode - 0xFF61 + 0xA1;
            if(jisvalue != 0)
            {
                wanted_shift = 2;
                bytes = 1;
            }
        }
        if(jisvalue == 0
        && unicode >= min_uni_jis0201lo && unicode <= max_uni_jis0201lo)
        {
            jisvalue = uni_jis0201lo[unicode-min_uni_jis0201lo];
            if(jisvalue != 0)
            {
                wanted_shift = 3;
                bytes = 1;
            }
        }

        if(shift != wanted_shift)
        {
            switch(wanted_shift)
            {
                case 0: result += "(B"; break;
                case 1: result += "$B"; break;
                case 2: result += "(I"; break;
                case 3: result += "(J"; break;
            }
            shift = wanted_shift;
        }
        
        if(bytes == 1)
        {
            char ch = jisvalue & 127;
            result += ch;
        }
        else
        {
            char ch1 = (jisvalue >> 8) & 255;
            char ch2 = (jisvalue     ) & 255;
            result += ch1;
            result += ch2;
        }
        if(bytes == 0) result += '?';
    }
    virtual void SeqEnd(std::string& result) const
    {
        result += "(B";
    }
    virtual bool Encodable(wchar_t ch, int shift) const
    {
        unsigned unicode = ch;
        if(unicode >= min_uni_euc && unicode <= max_uni_euc)
            if(uni_euc[unicode-min_uni_euc] != 0) return true;
        if(unicode >= 0xFF61 && unicode <= 0xFF9F) return true;

        /* Only offer to encode the low part of JIS0201 if
         * we're already in some JIS encoding *and* the character
         * is not ASCII.
         */
        if(shift >= 1 && unicode >= 0x100
        && unicode >= min_uni_jis0201lo && unicode <= max_uni_jis0201lo)
        {
            if(uni_jis0201lo[unicode-min_uni_jis0201lo] != 0) return true;
        }
        
        return false;
    }
    
    /* Reading */
    virtual const SeqResult IsShift(const std::string& input, unsigned pos) const
    {
        /* Shift code is at least 3 bytes long */
        if(pos+3 > input.size()) return false;
        if(!input.compare(pos, 3, "(B")) return SeqResult(0, 3);
        if(!input.compare(pos, 3, "$B")) return SeqResult(1, 3);
        if(!input.compare(pos, 3, "(I")) return SeqResult(2, 3);
        if(!input.compare(pos, 3, "(J")) return SeqResult(3, 3);
        return false;
    }
    virtual const SeqResult IsData(const std::string& input, unsigned pos, int shift) const
    {
        /* ISO-2022-JP only has data in shifted mode. */
        /* 7-bit ASCII is currently not handled here
         * to avoid stealing anyone's control codes.
         */
        switch(shift)
        {
            default:
            case 0: return false;
            case 1: // JIS0208
            {
                int b0 = (unsigned char)input[pos];
                int b1 = pos+1 < input.size() ? (unsigned char)input[pos+1] : EOL;
                
                if(b1 != EOL)
                {
                    unsigned value = (b0 << 8) | b1;
                    if(value >= min_euc && value <= max_euc)
                    {
                        value = euc_uni[value-min_euc];
                        if(value != 0) return SeqResult(true, 2, value);
                    }
                    /* We were in the shift, but there was unrecognized value! */
                    // To prevent terrible desyncing, we'll offer to decode it.
                    return SeqResult(true, 1, b0);
                }
                return false;
            }
            case 2: // JIS0201 HI (thin kana; actually the same as singlebyte SJIS)
            {
                int b0 = (unsigned char)input[pos];
                b0 |= 128;
                if(b0 < 0xA1 || b0 > 0xDF) return false;
                return SeqResult(true, 1, b0 - 0xA1 + 0xFF61);
            }
            case 3: // JIS0201 LO
            {
                int b0 = (unsigned char)input[pos];
                if(b0 < min_jis0201lo || b0 > max_jis0201lo) return false;
                unsigned value = jis0201lo_uni[b0 - min_jis0201lo];
                if(value != 0) return SeqResult(true, 1, value);
                /* We were in the shift, but there was unrecognized value! */
                // To prevent terrible desyncing, we'll offer to decode it.
                return SeqResult(true, 1, b0 & 127);
            }
        }
    }
};

class EUCJP: public Encoding
{
public:
    virtual ~EUCJP() {}
    
    virtual const Fondness GetFondness() const
    {
        /* This encoding can not actually be distinguished
         * from ISO-8859-*.
         */
        return VeryVague;
    }

    virtual void SeqValue(std::string& result, wchar_t ch, int&) const
    {
        unsigned jisvalue=0;
        unsigned unicode = ch;
        if(unicode >= min_uni_euc && unicode <= max_uni_euc)
            jisvalue = uni_euc[unicode-min_uni_euc];
        
        jisvalue |= 0x8080;
        
        char ch1 = (jisvalue >> 8) & 255;
        char ch2 = (jisvalue     ) & 255;
        result += ch1;
        result += ch2;
    }
    virtual void SeqEnd(std::string&) const
    {
        // nothing to do
    }
    virtual bool Encodable(wchar_t ch, int shift) const
    {
        unsigned unicode = ch;
        if(unicode >= min_uni_euc && unicode <= max_uni_euc)
            return uni_euc[unicode-min_uni_euc] != 0;
        return false;
    }
    
    /* Reading */
    virtual const SeqResult IsShift(const std::string&, unsigned) const
    {
        return false;
    }
    virtual const SeqResult IsData(const std::string& input, unsigned pos, int shift) const
    {
        if(shift != 1) return false;
        
        int b0 = (unsigned char)input[pos];
        int b1 = pos+1 < input.size() ? (unsigned char)input[pos+1] : EOL;
        
        if(b1 != EOL)
        {
            unsigned value = (b0 << 8) | b1;
            if((value & 0x8080) != 0x8080) return false;
            value &= 0x7F7F;
            if(value >= min_euc && value <= max_euc)
            {
                value = euc_uni[value-min_euc];
                if(value != 0) return SeqResult(true, 2, value);
            }
        }
        return false;
    }
};

class ShiftJIS: public Encoding
{
public:
    virtual ~ShiftJIS() {}
    
    virtual const Fondness GetFondness() const
    {
        /* Shift-JIS is very easily confused with other encodings.
         * We give it a "very vague" rating, which should give
         * ISO-8859-* priority over it when there are no control
         * characters in question.
         */
        return VeryVague;
    }

    virtual void SeqValue(std::string& result, wchar_t ch, int&) const
    {
        unsigned sjisvalue=0;
        unsigned unicode = ch;
        /* The thin kana letters are single-byte */
        if(unicode >= 0xFF61 && unicode <= 0xFF9F)
        {
            result += (char)(unicode - 0xFF61 + 0xA1);
            return;
        }
        if(unicode >= min_uni_sjis && unicode <= max_uni_sjis)
            sjisvalue = uni_sjis[unicode-min_uni_sjis];
        
        char ch1 = (sjisvalue >> 8) & 255;
        char ch2 = (sjisvalue     ) & 255;
        result += ch1;
        result += ch2;
    }
    virtual void SeqEnd(std::string&) const
    {
        // nothing to do
    }
    virtual bool Encodable(wchar_t ch, int shift) const
    {
        unsigned unicode = ch;
        if(unicode >= 0xFF61 && unicode <= 0xFF9F)
        {
            return true;
        }
        if(unicode >= min_uni_sjis && unicode <= max_uni_sjis)
            return uni_sjis[unicode-min_uni_sjis] != 0;
        return false;
    }
    
    /* Reading */
    virtual const SeqResult IsShift(const std::string&, unsigned) const
    {
        return false;
    }
    virtual const SeqResult IsData(const std::string& input, unsigned pos, int) const
    {
        int b0 = (unsigned char)input[pos];
        int b1 = pos+1 < input.size() ? (unsigned char)input[pos+1] : EOL;
        
        /* The thin kana letters are single-byte */
        if(b0 >= 0xA1 && b0 <= 0xDF)
        {
            return SeqResult(true, 1, b0 - 0xA1 + 0xFF61);
        }
        if(b1 != EOL)
        {
            unsigned value = (b0 << 8) | b1;
            if(value >= min_sjis && value <= max_sjis)
            {
                value = sjis_uni[value-min_sjis];
                if(value != 0) return SeqResult(true, 2, value);
            }
        }
        return false;
    }
};

const Encoding& FindEncoder(const std::string& name)
{
    std::string n = name;
    for(unsigned a=0; a<n.size(); ++a) n[a] = std::tolower(n[a]);
    
    if(n == "utf-8" || n == "utf8")
    {
        static const UTF8 utf8;
        return utf8;
    }
    if(n == "latin1" || n == "iso-8859-1" || n == "iso8859-1")
    {
        static const LatinEncoding latin1(uni_iso8859_1, iso8859_1_uni, min_uni_iso8859_1, max_uni_iso8859_1);
        return latin1;
    }
    if(n == "latin2" || n == "iso-8859-2" || n == "iso8859-2")
    {
        static const LatinEncoding latin2(uni_iso8859_2, iso8859_2_uni, min_uni_iso8859_2, max_uni_iso8859_2);
        return latin2;
    }
    if(n == "latin3" || n == "iso-8859-3" || n == "iso8859-3")
    {
        static const LatinEncoding latin3(uni_iso8859_3, iso8859_3_uni, min_uni_iso8859_3, max_uni_iso8859_3);
        return latin3;
    }
    if(n == "latin4" || n == "iso-8859-4" || n == "iso8859-4")
    {
        static const LatinEncoding latin4(uni_iso8859_4, iso8859_4_uni, min_uni_iso8859_4, max_uni_iso8859_4);
        return latin4;
    }
    if(n == "iso-8859-5" || n == "iso8859-5")
    {
        static const LatinEncoding latin5(uni_iso8859_5, iso8859_5_uni, min_uni_iso8859_5, max_uni_iso8859_5);
        return latin5;
    }
    if(n == "iso-8859-6" || n == "iso8859-6")
    {
        static const LatinEncoding latin6(uni_iso8859_6, iso8859_6_uni, min_uni_iso8859_6, max_uni_iso8859_6);
        return latin6;
    }
    if(n == "iso-8859-7" || n == "iso8859-7")
    {
        static const LatinEncoding latin7(uni_iso8859_7, iso8859_7_uni, min_uni_iso8859_7, max_uni_iso8859_7);
        return latin7;
    }
    if(n == "iso-8859-8" || n == "iso8859-8")
    {
        static const LatinEncoding latin8(uni_iso8859_8, iso8859_8_uni, min_uni_iso8859_8, max_uni_iso8859_8);
        return latin8;
    }
    if(n == "latin5" || n == "iso-8859-9" || n == "iso8859-9")
    {
        static const LatinEncoding latin9(uni_iso8859_9, iso8859_9_uni, min_uni_iso8859_9, max_uni_iso8859_9);
        return latin9;
    }
    if(n == "latin6" || n == "iso-8859-10" || n == "iso8859-10")
    {
        static const LatinEncoding latin10(uni_iso8859_10, iso8859_10_uni, min_uni_iso8859_10, max_uni_iso8859_10);
        return latin10;
    }
    if(n == "iso-8859-11" || n == "iso8859-11")
    {
        static const LatinEncoding latin11(uni_iso8859_11, iso8859_11_uni, min_uni_iso8859_11, max_uni_iso8859_11);
        return latin11;
    }
    if(n == "latin7" || n == "iso-8859-13" || n == "iso8859-13")
    {
        static const LatinEncoding latin13(uni_iso8859_13, iso8859_13_uni, min_uni_iso8859_13, max_uni_iso8859_13);
        return latin13;
    }
    if(n == "latin8" || n == "iso-8859-14" || n == "iso8859-14")
    {
        static const LatinEncoding latin14(uni_iso8859_14, iso8859_14_uni, min_uni_iso8859_14, max_uni_iso8859_14);
        return latin14;
    }
    if(n == "latin9" || n == "iso-8859-15" || n == "iso8859-15")
    {
        static const LatinEncoding latin15(uni_iso8859_15, iso8859_15_uni, min_uni_iso8859_15, max_uni_iso8859_15);
        return latin15;
    }
    if(n == "latin10" || n == "iso-8859-16" || n == "iso8859-16")
    {
        static const LatinEncoding latin16(uni_iso8859_16, iso8859_16_uni, min_uni_iso8859_16, max_uni_iso8859_16);
        return latin16;
    }
    if(n == "iso-2022-jp" || n == "iso2022jp" || n == "jis")
    {
        static const Iso2022JP iso2022jp;
        return iso2022jp;
    }
    if(n == "sjis" || n == "shift_jis" || n == "shiftjis" || n == "shift-jis")
    {
        static const ShiftJIS shiftjis;
        return shiftjis;
    }
    throw name;
}
